/**
 * This class is used to create a Client object.
 * The client handels the client server communication.
 * It communicates with the server (e.g. connection or query)
 * and recieves data from it, if dynamicMode is true.
 * If dynamicMode is false, the client can also handle data that is already
 * loaded and is not coming in in chunks.
 */
class Client {
  constructor({corpusId = null,
               socket = null,
               logging = true,
               dynamicMode = true} = {}) {
    this.corpusId = corpusId;
    this.dynamicMode = dynamicMode;
    this.logging = logging;
    this.requestQueryProgress = 0;
    this.socket = socket;
    this.eventListeners = {};
    this.connected = false;


    /**
     * Disable all console logging.
     * This is global. So every other log message in every other Class or
     * function used in conjunction with the client either logs or does not
     * log depending on the logging flag.
     * Credits to https://gist.github.com/kmonsoor/0244fdb4ad79a4826371e58a1a5fa984
     */
    if (!logging) {
      (() => {
        let console = (window.console = window.console || {});
        [
          'assert', 'clear', 'count', 'debug', 'dir', 'dirxml',
          'error', 'exception', 'group', 'groupCollapsed', 'groupEnd',
          'info', 'log', 'markTimeline', 'profile', 'profileEnd', 'table',
          'time', 'timeEnd', 'timeStamp', 'trace', 'warn'
        ].forEach((method) => {
          console[method] = () => {};
        });
      })();
    }
    console.info("Client initialized:", this);
  }

  // Registers one or more SocketEventListeners to the Client.
  setSocketEventListeners(eventListeners) {
    for (let eventListener of eventListeners) {
      this.eventListeners[eventListener.type] = eventListener;
    }
  }

  /**
   * Loads the SocketEventListeners so they will be triggered on their assigned
   * type strings because they double as the socket event event names.
   */
  loadSocketEventListeners() {
    for (let [type, listener] of Object.entries(this.eventListeners)) {
      listener.listenerFunction(type, this);
    }
  }

  /**
   * This functions sends events to the View to trigger specific functions that
   * are handleing the representation of data stored in the model.
   */
  notifyView(caseIdentifier, detailObject={}) {
    detailObject.caseIdentifier = caseIdentifier;
    const event = new CustomEvent('notify-view', { detail: detailObject });
    console.info('Client dispatching Notification:', caseIdentifier);
    document.dispatchEvent(event);
  }

  // Registers a CorpusAnalysisDisplay object to the Client.
  setDisplay(type, corpusAnalysisDisplay) {
    this.displays[type] = corpusAnalysisDisplay;
  }

  /**
   * Connects to the corpus analysis session for the specified corpus via
   * socket.io.
   */
  connect() {
    console.info('corpus_analysis_init: Client connecting to session via',
                 'socket.emit');
    this.socket.emit('corpus_analysis_init', this.corpusId);
  }

  getMetaData() {
    console.info('corpus_analysis_meta_data: Client getting meta data via',
                 'socket.emit.');
    this.socket.emit('corpus_analysis_meta_data', this.corpusId);
  }

  /**
   * Emits query to the server via socket.io. Server will send the results
   * back.
   */
  query(queryStr) {
    console.info('corpus_analysis_query: Client sending query via',
                 'socket.emit for the query', queryStr);
    this.socket.emit('corpus_analysis_query', queryStr);
  }

  /**
   * Create results data either from all results or from all marked sub results
   * Triggers emit to get full match context from server for a number of
   * matches identified by their data_index.
   **/
  getResultsData(resultsType, dataIndexes, results) {
    let tmp_first_cpos = [];
    let tmp_last_cpos = [];
    for (let dataIndex of dataIndexes) {
      tmp_first_cpos.push(results.data.matches[dataIndex].c[0]);
      tmp_last_cpos.push(results.data.matches[dataIndex].c[1]);
    }
    nopaque.socket.emit("corpus_analysis_inspect_match",
                        {type: resultsType,
                         data_indexes: dataIndexes,
                         first_cpos: tmp_first_cpos,
                         last_cpos: tmp_last_cpos,});
  }

}


/**
 * This class is used to create an SocketEventListener.
 * Input are an identifying type string, the listener function and callbacks
 * which will be executed as part of the listener function. The identifying
 * type string is also used as the socket event event identifier.
 */
class ClientEventListener {
  constructor(type, listenerFunction) {
    this.listenerCallbacks = {};
    this.listenerFunction = listenerFunction;
    this.type = type;
  }

  // Registers callbacks to this SocketEventListener
  setCallbacks(listenerCallbacks) {
    for (let listenerCallback of listenerCallbacks) {
      this.listenerCallbacks[listenerCallback.type] = listenerCallback;
    }
  }

  /** Shorthand to execute all registered callbacks with same defaultArgs
   * in insertion order.
   * NOTE:
   * Since ECMAScript 2015, objects do preserve creation order for
   * string and Symbol keys. In JavaScript engines that comply with the
   * ECMAScript 2015 spec, iterating over an object with only string keys will
   * yield the keys in order of insertion.
   * So all modern Browsers.
   */
  executeCallbacks(defaultArgs) {
    for (let [type, listenerCallback] of Object.entries(this.listenerCallbacks)) {
      listenerCallback.callbackFunction(...defaultArgs,
                                        ...listenerCallback.args);
    }
  }
  // use this if you only want to execute a specific registered callback
  executeCallback(defaultArgs, type) {
    let listenerCallback = this.listenerCallbacks[type];
    listenerCallback.callbackFunction(...defaultArgs,
                                      ...listenerCallback.args);
  }
}

/**
 * This class is used to create an ListenerCallback which will be registered
 * to an SocketEventListener so the Listener can invoke the ListenerCallback
 * callback functions.
 */
class ListenerCallback {
  constructor(type, callbackFunction, argsList) {
    this.args = argsList;
    this.callbackFunction = callbackFunction;
    this.type = type;
  }
}

// export Classes from this module
export {
  Client,
  ClientEventListener,
  ListenerCallback,
};